// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information
use crate::{
    array::{Array, ArrayType},
    bundle::{BundleType, BundleValue, DynBundle, DynBundleType, FieldType},
    clock::{Clock, ClockType},
    enum_::{DynEnum, DynEnumType, EnumType, VariantType},
    expr::{Expr, ToExpr},
    int::{DynIntType, DynSInt, DynSIntType, DynUInt, DynUIntType, IntTypeTrait, IntValue},
    intern::{Intern, Interned, InternedCompare, PtrEqWithTypeId, SupportsPtrEqWithTypeId},
    module::{
        transform::visit::{Fold, Folder, Visit, Visitor},
        ModuleBuilder, NormalModule,
    },
    reset::{AsyncReset, AsyncResetType, Reset, ResetType, SyncReset, SyncResetType},
    source_location::SourceLocation,
    util::{iter_eq_by, GenericConstBool},
    valueless::Valueless,
};
use std::{
    any::{Any, TypeId},
    convert::Infallible,
    fmt,
    hash::{Hash, Hasher},
    iter::FusedIterator,
    marker::PhantomData,
    ops::Deref,
    sync::Arc,
};

#[doc(inline)]
pub use fayalite_proc_macros::Value;

mod sealed {
    pub trait Sealed {}
}

#[derive(Copy, Clone, Eq, PartialEq, Hash, Debug)]
pub enum TypeEnum {
    BundleType(DynBundleType),
    EnumType(DynEnumType),
    ArrayType(ArrayType<[DynCanonicalValue]>),
    UInt(DynUIntType),
    SInt(DynSIntType),
    Clock(ClockType),
    AsyncReset(AsyncResetType),
    SyncReset(SyncResetType),
    Reset(ResetType),
}

#[derive(Copy, Clone, Eq, PartialEq, Hash, Debug)]
pub enum CanonicalTypeKind {
    BundleType,
    EnumType,
    ArrayType,
    UInt,
    SInt,
    Clock,
    AsyncReset,
    SyncReset,
    Reset,
    DynCanonicalType,
}

impl CanonicalTypeKind {
    pub fn expr_ty(self) -> &'static str {
        match self {
            CanonicalTypeKind::BundleType => "Expr<DynBundle>",
            CanonicalTypeKind::EnumType => "Expr<DynEnum>",
            CanonicalTypeKind::ArrayType => "Expr<Array<[DynCanonicalValue]>>",
            CanonicalTypeKind::UInt => "Expr<DynUInt>",
            CanonicalTypeKind::SInt => "Expr<DynSInt>",
            CanonicalTypeKind::Clock => "Expr<Clock>",
            CanonicalTypeKind::AsyncReset => "Expr<AsyncReset>",
            CanonicalTypeKind::SyncReset => "Expr<SyncReset>",
            CanonicalTypeKind::Reset => "Expr<Reset>",
            CanonicalTypeKind::DynCanonicalType => "Expr<DynCanonicalValue>",
        }
    }
}

#[derive(Clone, Eq, PartialEq, Hash, Debug)]
pub enum ValueEnum {
    Bundle(DynBundle),
    Enum(DynEnum),
    Array(Array<[DynCanonicalValue]>),
    UInt(DynUInt),
    SInt(DynSInt),
    Clock(Clock),
    AsyncReset(AsyncReset),
    SyncReset(SyncReset),
    Reset(Reset),
}

impl TypeEnum {
    pub fn is_passive(self) -> bool {
        match self {
            TypeEnum::BundleType(ty) => DynBundleType::is_passive(ty),
            TypeEnum::EnumType(ty) => DynEnumType::is_passive(ty),
            TypeEnum::ArrayType(ty) => ty.element().is_passive(),
            TypeEnum::UInt(_) => true,
            TypeEnum::SInt(_) => true,
            TypeEnum::Clock(_) => true,
            TypeEnum::AsyncReset(_) => true,
            TypeEnum::SyncReset(_) => true,
            TypeEnum::Reset(_) => true,
        }
    }
    pub fn is_storable(self) -> bool {
        match self {
            TypeEnum::BundleType(ty) => DynBundleType::is_storable(ty),
            TypeEnum::EnumType(ty) => DynEnumType::is_storable(ty),
            TypeEnum::ArrayType(ty) => ty.element().is_storable(),
            TypeEnum::UInt(_) => true,
            TypeEnum::SInt(_) => true,
            TypeEnum::Clock(_) => false,
            TypeEnum::AsyncReset(_) => false,
            TypeEnum::SyncReset(_) => false,
            TypeEnum::Reset(_) => false,
        }
    }
    pub fn is_castable_from_bits(self) -> bool {
        match self {
            TypeEnum::BundleType(ty) => DynBundleType::is_castable_from_bits(ty),
            TypeEnum::EnumType(ty) => DynEnumType::is_castable_from_bits(ty),
            TypeEnum::ArrayType(ty) => ty.element().is_castable_from_bits(),
            TypeEnum::UInt(_) => true,
            TypeEnum::SInt(_) => true,
            TypeEnum::Clock(_) => true,
            TypeEnum::AsyncReset(_) => true,
            TypeEnum::SyncReset(_) => true,
            TypeEnum::Reset(_) => false, // Reset is not castable from bits because we don't know if it's async or sync
        }
    }
    pub fn bit_width(self) -> usize {
        match self {
            TypeEnum::BundleType(ty) => DynBundleType::bit_width(ty),
            TypeEnum::EnumType(ty) => DynEnumType::bit_width(ty),
            TypeEnum::ArrayType(ty) => ArrayType::bit_width(&ty),
            TypeEnum::UInt(ty) => ty.width,
            TypeEnum::SInt(ty) => ty.width,
            TypeEnum::Clock(ClockType)
            | TypeEnum::AsyncReset(AsyncResetType)
            | TypeEnum::SyncReset(SyncResetType)
            | TypeEnum::Reset(ResetType) => 1,
        }
    }
    pub fn bundle_type(self) -> Option<DynBundleType> {
        if let TypeEnum::BundleType(retval) = self {
            Some(retval)
        } else {
            None
        }
    }
    pub fn enum_type(self) -> Option<DynEnumType> {
        if let TypeEnum::EnumType(retval) = self {
            Some(retval)
        } else {
            None
        }
    }
    pub fn array_type(self) -> Option<ArrayType<[DynCanonicalValue]>> {
        if let TypeEnum::ArrayType(retval) = self {
            Some(retval)
        } else {
            None
        }
    }
    pub fn uint(self) -> Option<DynUIntType> {
        if let TypeEnum::UInt(retval) = self {
            Some(retval)
        } else {
            None
        }
    }
    pub fn sint(self) -> Option<DynSIntType> {
        if let TypeEnum::SInt(retval) = self {
            Some(retval)
        } else {
            None
        }
    }
    pub fn clock(self) -> Option<ClockType> {
        if let TypeEnum::Clock(retval) = self {
            Some(retval)
        } else {
            None
        }
    }
    pub fn async_reset(self) -> Option<AsyncResetType> {
        if let TypeEnum::AsyncReset(retval) = self {
            Some(retval)
        } else {
            None
        }
    }
    pub fn sync_reset(self) -> Option<SyncResetType> {
        if let TypeEnum::SyncReset(retval) = self {
            Some(retval)
        } else {
            None
        }
    }
    pub fn reset(self) -> Option<ResetType> {
        if let TypeEnum::Reset(retval) = self {
            Some(retval)
        } else {
            None
        }
    }
    pub fn to_dyn(self) -> Interned<dyn DynType> {
        match self {
            TypeEnum::BundleType(ty) => ty.to_dyn(),
            TypeEnum::EnumType(ty) => ty.to_dyn(),
            TypeEnum::ArrayType(ty) => ty.to_dyn(),
            TypeEnum::UInt(ty) => ty.to_dyn(),
            TypeEnum::SInt(ty) => ty.to_dyn(),
            TypeEnum::Clock(ty) => ty.to_dyn(),
            TypeEnum::AsyncReset(ty) => ty.to_dyn(),
            TypeEnum::SyncReset(ty) => ty.to_dyn(),
            TypeEnum::Reset(ty) => ty.to_dyn(),
        }
    }
    pub fn canonical_dyn(self) -> Interned<dyn DynCanonicalType> {
        match self {
            TypeEnum::BundleType(ty) => ty.canonical_dyn(),
            TypeEnum::EnumType(ty) => ty.canonical_dyn(),
            TypeEnum::ArrayType(ty) => ty.canonical_dyn(),
            TypeEnum::UInt(ty) => ty.canonical_dyn(),
            TypeEnum::SInt(ty) => ty.canonical_dyn(),
            TypeEnum::Clock(ty) => ty.canonical_dyn(),
            TypeEnum::AsyncReset(ty) => ty.canonical_dyn(),
            TypeEnum::SyncReset(ty) => ty.canonical_dyn(),
            TypeEnum::Reset(ty) => ty.canonical_dyn(),
        }
    }
}

pub trait DynType: fmt::Debug + Send + Sync + Any + SupportsPtrEqWithTypeId {
    fn canonical_dyn(&self) -> Interned<dyn DynCanonicalType>;
    fn to_dyn(&self) -> Interned<dyn DynType>;
    fn mask_type_dyn(&self) -> Interned<dyn DynType>;
    fn source_location_dyn(&self) -> SourceLocation;
    fn type_enum_dyn(&self) -> TypeEnum;
    fn is_passive(&self) -> bool;
    fn is_storable(&self) -> bool;
    fn is_castable_from_bits(&self) -> bool;
    fn bit_width(&self) -> usize;
    fn as_dyn_type(&self) -> &dyn DynType;
    fn as_dyn_canonical_type(&self) -> Option<&dyn DynCanonicalType>;
    #[deprecated = "use <dyn DynType>::downcast or <dyn DynCanonicalType>::downcast instead, they properly handle Interned<dyn DynType> and Interned<dyn DynCanonicalType>"]
    #[doc(hidden)]
    fn as_any(&self) -> &dyn Any;
    fn type_name_for_display_only(&self) -> &'static str;
    fn from_dyn_canonical_type(ty: Interned<dyn DynCanonicalType>) -> Self
    where
        Self: Type,
    {
        Self::from_canonical_type(
            <dyn DynCanonicalType>::downcast::<Self::CanonicalType>(&*ty)
                .expect("type should match"),
        )
    }
}

impl InternedCompare for dyn DynType {
    type InternedCompareKey = PtrEqWithTypeId;
    fn interned_compare_key_ref(this: &Self) -> Self::InternedCompareKey {
        Self::get_ptr_eq_with_type_id(this)
    }
    fn interned_compare_key_weak(this: &std::sync::Weak<Self>) -> Self::InternedCompareKey {
        Self::get_ptr_eq_with_type_id(&*this.upgrade().unwrap())
    }
}

impl<T: Type> DynType for T {
    fn canonical_dyn(&self) -> Interned<dyn DynCanonicalType> {
        T::canonical_dyn_impl(self)
    }

    fn to_dyn(&self) -> Interned<dyn DynType> {
        T::to_dyn_impl(self)
    }

    fn mask_type_dyn(&self) -> Interned<dyn DynType> {
        T::mask_type(self).to_dyn()
    }

    fn source_location_dyn(&self) -> SourceLocation {
        self.source_location()
    }

    fn type_enum_dyn(&self) -> TypeEnum {
        self.type_enum()
    }

    fn is_passive(&self) -> bool {
        self.type_enum().is_passive()
    }

    fn is_storable(&self) -> bool {
        self.type_enum().is_storable()
    }

    fn is_castable_from_bits(&self) -> bool {
        self.type_enum().is_castable_from_bits()
    }

    fn bit_width(&self) -> usize {
        self.type_enum().bit_width()
    }

    fn as_dyn_type(&self) -> &dyn DynType {
        self
    }

    fn as_any(&self) -> &dyn Any {
        self
    }

    fn as_dyn_canonical_type(&self) -> Option<&dyn DynCanonicalType> {
        T::as_dyn_canonical_type_impl(self)
    }

    fn type_name_for_display_only(&self) -> &'static str {
        std::any::type_name::<Self>()
    }
}

impl<State: ?Sized + Visitor> Visit<State> for Interned<dyn DynType> {
    fn visit(&self, state: &mut State) -> Result<(), State::Error> {
        state.visit_interned_dyn_type(self)
    }
    fn default_visit(&self, state: &mut State) -> Result<(), State::Error> {
        self.type_enum().visit(state)
    }
}

impl<State: ?Sized + Folder> Fold<State> for Interned<dyn DynType> {
    fn fold(self, state: &mut State) -> Result<Self, <State>::Error> {
        state.fold_interned_dyn_type(self)
    }

    fn default_fold(self, state: &mut State) -> Result<Self, <State>::Error> {
        Ok(self.type_enum().fold(state)?.to_dyn())
    }
}

impl<State: ?Sized + Visitor> Visit<State> for Interned<dyn DynCanonicalType> {
    fn visit(&self, state: &mut State) -> Result<(), State::Error> {
        state.visit_interned_dyn_canonical_type(self)
    }
    fn default_visit(&self, state: &mut State) -> Result<(), State::Error> {
        self.type_enum().visit(state)
    }
}

impl<State: ?Sized + Folder> Fold<State> for Interned<dyn DynCanonicalType> {
    fn fold(self, state: &mut State) -> Result<Self, <State>::Error> {
        state.fold_interned_dyn_canonical_type(self)
    }

    fn default_fold(self, state: &mut State) -> Result<Self, <State>::Error> {
        Ok(self.type_enum().fold(state)?.canonical_dyn())
    }
}

#[derive(Clone)]
enum DowncastFailedReason<'a, T: Type> {
    NotACanonicalType { dyn_type: &'a dyn DynType },
    TypeIdMismatch { dyn_type: &'a dyn DynType },
    _Unused(Infallible, PhantomData<fn() -> T>),
}

#[derive(Clone)]
pub struct DowncastFailed<'a, T: Type>(DowncastFailedReason<'a, T>);

impl<T: Type> fmt::Debug for DowncastFailed<'_, T> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self.0 {
            DowncastFailedReason::NotACanonicalType { dyn_type } => f
                .debug_struct("DowncastFailed::NotACanonicalType")
                .field("dyn_type", &dyn_type)
                .field("target_type", &std::any::type_name::<T>())
                .finish(),
            #[allow(deprecated)]
            DowncastFailedReason::TypeIdMismatch { dyn_type } => f
                .debug_struct("DowncastFailed::TypeIdMismatch")
                .field("dyn_type", &dyn_type)
                .field("dyn_type_name", &dyn_type.type_name_for_display_only())
                .field("dyn_type_id", &dyn_type.as_any().type_id())
                .field("target_type", &std::any::type_name::<T>())
                .field("target_type_id", &TypeId::of::<T>())
                .finish(),
            DowncastFailedReason::_Unused(v, _) => match v {},
        }
    }
}

impl dyn DynType {
    pub fn downcast<T: Type>(&self) -> Result<T, DowncastFailed<T>> {
        let mut this = self;
        loop {
            #[allow(deprecated)]
            if let Some(v) = this.as_any().downcast_ref::<Interned<dyn DynType>>() {
                this = &**v;
            } else if let Some(v) = this
                .as_any()
                .downcast_ref::<Interned<dyn DynCanonicalType>>()
            {
                this = (**v).as_dyn_type();
            } else {
                break;
            }
        }
        if TypeId::of::<T>() == TypeId::of::<Interned<dyn DynType>>() {
            <dyn Any>::downcast_ref(&this.to_dyn()).cloned()
        } else if TypeId::of::<T>() == TypeId::of::<Interned<dyn DynCanonicalType>>() {
            let Some(dyn_canonical_type) = this.as_dyn_canonical_type() else {
                return Err(DowncastFailed(DowncastFailedReason::NotACanonicalType {
                    dyn_type: this,
                }));
            };
            let retval = dyn_canonical_type.canonical_dyn();
            <dyn Any>::downcast_ref(&retval).cloned()
        } else {
            #[allow(deprecated)]
            this.as_any().downcast_ref().cloned()
        }
        .ok_or(DowncastFailed(DowncastFailedReason::TypeIdMismatch {
            dyn_type: this,
        }))
    }
}

impl dyn DynCanonicalType {
    pub fn downcast<T: CanonicalType>(&self) -> Result<T, DowncastFailed<T>> {
        self.as_dyn_type().downcast()
    }
}

pub trait Connect<Rhs> {}

pub trait MatchVariantAndInactiveScope: Sized {
    type MatchVariant;
    type MatchActiveScope;
    fn match_activate_scope(self) -> (Self::MatchVariant, Self::MatchActiveScope);
}

#[derive(Debug)]
pub struct MatchVariantWithoutScope<T: 'static + Send + Sync>(pub T);

impl<T: 'static + Send + Sync> MatchVariantAndInactiveScope for MatchVariantWithoutScope<T> {
    type MatchVariant = T;
    type MatchActiveScope = ();

    fn match_activate_scope(self) -> (Self::MatchVariant, Self::MatchActiveScope) {
        (self.0, ())
    }
}

pub trait Type: DynType + Clone + Hash + Eq + Intern + Connect<Self> {
    type CanonicalType: CanonicalType<
        CanonicalType = Self::CanonicalType,
        CanonicalValue = Self::CanonicalValue,
        Value = Self::CanonicalValue,
    >;
    type Value: Value<Type = Self>;
    type CanonicalValue: CanonicalValue<Type = Self::CanonicalType>;
    type MaskType: Type<
        Value = Self::MaskValue,
        MaskType = Self::MaskType,
        MaskValue = Self::MaskValue,
    >;
    type MaskValue: Value<Type = Self::MaskType>;
    type MatchVariant: 'static + Send + Sync;
    type MatchActiveScope;
    type MatchVariantAndInactiveScope: MatchVariantAndInactiveScope<
        MatchVariant = Self::MatchVariant,
        MatchActiveScope = Self::MatchActiveScope,
    >;
    type MatchVariantsIter: Iterator<Item = Self::MatchVariantAndInactiveScope>
        + ExactSizeIterator
        + FusedIterator
        + DoubleEndedIterator;
    fn match_variants<IO: BundleValue>(
        this: Expr<Self::Value>,
        module_builder: &mut ModuleBuilder<IO, NormalModule>,
        source_location: SourceLocation,
    ) -> Self::MatchVariantsIter
    where
        IO::Type: BundleType<Value = IO>;
    fn mask_type(&self) -> Self::MaskType;
    fn canonical(&self) -> Self::CanonicalType;
    fn canonical_dyn_impl(this: &Self) -> Interned<dyn DynCanonicalType> {
        Interned::cast_unchecked(this.canonical().intern(), |v| -> &dyn DynCanonicalType {
            v
        })
    }
    fn to_dyn_impl(this: &Self) -> Interned<dyn DynType> {
        Interned::cast_unchecked(this.intern(), |v| -> &dyn DynType { v })
    }
    fn as_dyn_canonical_type_impl(this: &Self) -> Option<&dyn DynCanonicalType> {
        let _ = this;
        None
    }
    fn source_location(&self) -> SourceLocation;
    fn type_enum(&self) -> TypeEnum;
    fn from_canonical_type(t: Self::CanonicalType) -> Self;
}

pub trait TypeWithDeref: Type {
    fn expr_deref(this: &Expr<Self::Value>) -> &Self::MatchVariant;
}

macro_rules! impl_match_values_as_self {
    () => {
        type MatchVariant = crate::expr::Expr<Self::Value>;
        type MatchActiveScope = ();
        type MatchVariantAndInactiveScope = crate::ty::MatchVariantWithoutScope<Self::MatchVariant>;
        type MatchVariantsIter = std::iter::Once<Self::MatchVariantAndInactiveScope>;

        fn match_variants<IO: crate::bundle::BundleValue>(
            this: crate::expr::Expr<Self::Value>,
            module_builder: &mut crate::module::ModuleBuilder<IO, crate::module::NormalModule>,
            source_location: crate::source_location::SourceLocation,
        ) -> Self::MatchVariantsIter
        where
            IO::Type: crate::bundle::BundleType<Value = IO>,
        {
            let _ = module_builder;
            let _ = source_location;
            std::iter::once(crate::ty::MatchVariantWithoutScope(this))
        }
    };
}

use bitvec::slice::BitSlice;
pub(crate) use impl_match_values_as_self;

pub trait DynCanonicalType: DynType + sealed::Sealed {}

impl InternedCompare for dyn DynCanonicalType {
    type InternedCompareKey = PtrEqWithTypeId;
    fn interned_compare_key_ref(this: &Self) -> Self::InternedCompareKey {
        Self::get_ptr_eq_with_type_id(this)
    }
    fn interned_compare_key_weak(this: &std::sync::Weak<Self>) -> Self::InternedCompareKey {
        Self::get_ptr_eq_with_type_id(&*this.upgrade().unwrap())
    }
}

pub trait CanonicalType:
    Copy
    + DynCanonicalType
    + Type<CanonicalType = Self, Value = <Self as Type>::CanonicalValue>
    + sealed::Sealed
    + Intern
    + Connect<Self>
{
    const CANONICAL_TYPE_KIND: CanonicalTypeKind;
    fn can_connect(&self, other: &Self) -> bool {
        macro_rules! unwrap_other {
            ($var:ident = $fn:ident()) => {
                let Some($var) = other.type_enum().$fn() else {
                    return false;
                };
            };
        }
        match self.type_enum() {
            TypeEnum::BundleType(this) => {
                unwrap_other!(other = bundle_type());
                iter_eq_by(
                    this.fields().iter().copied(),
                    other.fields().iter().copied(),
                    |this, other| -> bool {
                        let FieldType { name, flipped, ty } = this;
                        name == other.name && flipped == other.flipped && ty.can_connect(&other.ty)
                    },
                )
            }
            TypeEnum::EnumType(this) => {
                unwrap_other!(other = enum_type());
                iter_eq_by(
                    this.variants().iter().copied(),
                    other.variants().iter().copied(),
                    |this, other| -> bool {
                        let VariantType { name, ty } = this;
                        name == other.name
                            && iter_eq_by(ty, other.ty, |ty, other_ty| ty.can_connect(&other_ty))
                    },
                )
            }
            TypeEnum::ArrayType(this) => {
                unwrap_other!(other = array_type());
                this.len() == other.len() && this.element().can_connect(other.element())
            }
            TypeEnum::UInt(_this) => {
                unwrap_other!(_other = uint());
                true // all uint widths can connect to all other uint widths
            }
            TypeEnum::SInt(_this) => {
                unwrap_other!(_other = sint());
                true // all sint widths can connect to all other sint widths
            }
            TypeEnum::Clock(this) => {
                unwrap_other!(other = clock());
                this == other
            }
            TypeEnum::AsyncReset(this) => {
                unwrap_other!(other = async_reset());
                this == other
            }
            TypeEnum::SyncReset(this) => {
                unwrap_other!(other = sync_reset());
                this == other
            }
            TypeEnum::Reset(this) => {
                unwrap_other!(other = reset());
                this == other
            }
        }
    }
}

impl<T: CanonicalType> DynCanonicalType for T {}

pub trait StaticType: Type {
    fn static_type() -> Self;
}

pub trait DynValueTrait: fmt::Debug + Send + Sync + Any {
    fn ty_dyn(&self) -> Interned<dyn DynType>;
    fn to_canonical_dyn(&self) -> DynCanonicalValue;
    fn eq_dyn(&self, other: &dyn DynValueTrait) -> bool;
    fn hash_dyn(&self, hasher: &mut dyn Hasher);
    fn as_any(&self) -> &(dyn Any + Send + Sync);
    fn as_dyn_value_trait(&self) -> &dyn DynValueTrait;
    fn as_arc_dyn_value_trait(self: Arc<Self>) -> Arc<dyn DynValueTrait>;
    fn to_bits(&self) -> Interned<BitSlice>;
}

macro_rules! dyn_value {
    ($name:ident, trait = $trait:ident) => {
        #[derive(Clone)]
        pub struct $name(pub Arc<dyn $trait>);

        impl Hash for $name {
            fn hash<H: Hasher>(&self, state: &mut H) {
                self.0.hash_dyn(state);
            }
        }

        impl Eq for $name {}

        impl fmt::Debug for $name {
            fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
                self.0.fmt(f)
            }
        }

        impl Deref for $name {
            type Target = dyn $trait;

            fn deref(&self) -> &Self::Target {
                &*self.0
            }
        }
    };
}

dyn_value!(DynValue, trait = DynValueTrait);
dyn_value!(DynCanonicalValue, trait = DynCanonicalValueTrait);

impl PartialEq for DynValue {
    fn eq(&self, other: &Self) -> bool {
        self.0.eq_dyn(&*other.0)
    }
}

impl ToExpr for DynValue {
    type Type = Interned<dyn DynType>;

    fn ty(&self) -> Self::Type {
        self.0.ty_dyn()
    }

    fn to_expr(&self) -> Expr<Self> {
        Expr::from_value(self)
    }
}

impl Value for DynValue {
    fn ty_dyn_impl(this: &Self) -> Interned<dyn DynType> {
        this.ty()
    }

    fn to_canonical_dyn_impl(this: &Self) -> DynCanonicalValue {
        this.to_canonical()
    }

    fn to_canonical(&self) -> <Self::Type as Type>::CanonicalValue {
        self.0.to_canonical_dyn()
    }

    fn to_bits_impl(this: &Self) -> Interned<BitSlice> {
        this.0.to_bits()
    }
}

impl PartialEq for DynCanonicalValue {
    fn eq(&self, other: &Self) -> bool {
        self.0.eq_dyn(other.0.as_dyn_value_trait())
    }
}

impl ToExpr for DynCanonicalValue {
    type Type = Interned<dyn DynCanonicalType>;

    fn ty(&self) -> Self::Type {
        self.0.ty_dyn_canonical()
    }

    fn to_expr(&self) -> Expr<Self> {
        Expr::from_value(self)
    }
}

impl Value for DynCanonicalValue {
    fn ty_dyn_impl(this: &Self) -> Interned<dyn DynType> {
        this.0.ty_dyn()
    }

    fn to_canonical_dyn_impl(this: &Self) -> DynCanonicalValue {
        this.to_canonical()
    }

    fn to_canonical(&self) -> <Self::Type as Type>::CanonicalValue {
        self.0.to_canonical_dyn()
    }

    fn to_bits_impl(this: &Self) -> Interned<BitSlice> {
        <Self as CanonicalValue>::to_bits_impl(this)
    }
}

impl sealed::Sealed for DynCanonicalValue {}

impl CanonicalValue for DynCanonicalValue {
    fn ty_dyn_canonical_impl(this: &Self) -> Interned<dyn DynCanonicalType> {
        this.0.ty_dyn_canonical()
    }

    fn value_enum_impl(this: &Self) -> ValueEnum {
        this.0.value_enum()
    }

    fn to_bits_impl(this: &Self) -> Interned<BitSlice> {
        this.0.to_bits()
    }
}

impl<State: ?Sized + Folder> Fold<State> for DynCanonicalValue {
    fn fold(self, state: &mut State) -> Result<Self, <State as Folder>::Error> {
        state.fold_dyn_canonical_value(self)
    }
    fn default_fold(self, state: &mut State) -> Result<Self, <State as Folder>::Error> {
        Ok(match self.value_enum().fold(state)? {
            ValueEnum::Bundle(v) => v.to_canonical_dyn(),
            ValueEnum::Enum(v) => v.to_canonical_dyn(),
            ValueEnum::Array(v) => v.to_canonical_dyn(),
            ValueEnum::UInt(v) => v.to_canonical_dyn(),
            ValueEnum::SInt(v) => v.to_canonical_dyn(),
            ValueEnum::Clock(v) => v.to_canonical_dyn(),
            ValueEnum::AsyncReset(v) => v.to_canonical_dyn(),
            ValueEnum::SyncReset(v) => v.to_canonical_dyn(),
            ValueEnum::Reset(v) => v.to_canonical_dyn(),
        })
    }
}

impl<State: ?Sized + Visitor> Visit<State> for DynCanonicalValue {
    fn visit(&self, state: &mut State) -> Result<(), <State as Visitor>::Error> {
        state.visit_dyn_canonical_value(self)
    }
    fn default_visit(&self, state: &mut State) -> Result<(), <State as Visitor>::Error> {
        self.value_enum().visit(state)
    }
}

pub type AsMask<V> = <<V as ToExpr>::Type as Type>::MaskValue;

pub trait Value: DynValueTrait + Clone + Eq + Hash + ToExpr {
    fn ty_dyn_impl(this: &Self) -> Interned<dyn DynType> {
        Interned::cast_unchecked(this.ty().intern_sized(), |v: &Self::Type| -> &dyn DynType {
            v
        })
    }
    fn to_canonical(&self) -> <Self::Type as Type>::CanonicalValue;
    fn to_canonical_dyn_impl(this: &Self) -> DynCanonicalValue {
        DynCanonicalValue(Arc::new(this.to_canonical()))
    }
    fn valueless(&self) -> Valueless<Self::Type> {
        Valueless { ty: self.ty() }
    }
    fn to_bits_impl(this: &Self) -> Interned<BitSlice> {
        <<Self::Type as Type>::CanonicalValue as CanonicalValue>::to_bits_impl(&this.to_canonical())
    }
}

pub trait DynCanonicalValueTrait: DynValueTrait + sealed::Sealed {
    fn ty_dyn_canonical(&self) -> Interned<dyn DynCanonicalType>;
    fn value_enum(&self) -> ValueEnum;
}

impl<T: CanonicalValue> DynCanonicalValueTrait for T
where
    T::Type: CanonicalType<CanonicalType = T::Type, Value = Self, CanonicalValue = Self>,
{
    fn ty_dyn_canonical(&self) -> Interned<dyn DynCanonicalType> {
        Self::ty_dyn_canonical_impl(self)
    }
    fn value_enum(&self) -> ValueEnum {
        Self::value_enum_impl(self)
    }
}

pub trait CanonicalValue: DynCanonicalValueTrait + Value + sealed::Sealed
where
    Self::Type: CanonicalType<CanonicalType = Self::Type, Value = Self, CanonicalValue = Self>,
{
    fn ty_dyn_canonical_impl(this: &Self) -> Interned<dyn DynCanonicalType> {
        Interned::cast_unchecked(
            this.ty().intern_sized(),
            |v: &Self::Type| -> &dyn DynCanonicalType { v },
        )
    }
    fn value_enum_impl(this: &Self) -> ValueEnum;
    fn to_bits_impl(this: &Self) -> Interned<BitSlice>;
}

impl<T: Value> DynValueTrait for T
where
    <Self as ToExpr>::Type: Type<Value = Self>,
{
    fn ty_dyn(&self) -> Interned<dyn DynType> {
        T::ty_dyn_impl(self)
    }

    fn to_canonical_dyn(&self) -> DynCanonicalValue {
        T::to_canonical_dyn_impl(self)
    }

    fn eq_dyn(&self, other: &dyn DynValueTrait) -> bool {
        let other_any = other.as_any();
        if let Some(other) = other_any.downcast_ref::<Self>() {
            T::eq(self, other)
        } else {
            false
        }
    }

    fn hash_dyn(&self, mut hasher: &mut dyn Hasher) {
        self.hash(&mut hasher);
    }

    fn as_any(&self) -> &(dyn Any + Send + Sync) {
        self
    }

    fn as_dyn_value_trait(&self) -> &dyn DynValueTrait {
        self
    }

    fn as_arc_dyn_value_trait(self: Arc<Self>) -> Arc<dyn DynValueTrait> {
        self
    }

    fn to_bits(&self) -> Interned<BitSlice> {
        T::to_bits_impl(self)
    }
}

impl Type for Interned<dyn DynType> {
    type CanonicalType = Interned<dyn DynCanonicalType>;
    type Value = DynValue;
    type CanonicalValue = DynCanonicalValue;
    type MaskType = Interned<dyn DynType>;
    type MaskValue = DynValue;

    impl_match_values_as_self!();

    fn mask_type(&self) -> Self::MaskType {
        (**self).mask_type_dyn()
    }

    fn canonical(&self) -> Self::CanonicalType {
        (**self).canonical_dyn()
    }

    fn source_location(&self) -> SourceLocation {
        (**self).source_location_dyn()
    }

    fn type_enum(&self) -> TypeEnum {
        (**self).type_enum_dyn()
    }

    fn from_canonical_type(t: Self::CanonicalType) -> Self {
        Interned::cast_unchecked(t, <dyn DynCanonicalType>::as_dyn_type)
    }

    fn canonical_dyn_impl(this: &Self) -> Interned<dyn DynCanonicalType> {
        this.canonical()
    }

    fn to_dyn_impl(this: &Self) -> Interned<dyn DynType> {
        *this
    }

    fn as_dyn_canonical_type_impl(this: &Self) -> Option<&dyn DynCanonicalType> {
        (**this).as_dyn_canonical_type()
    }
}

impl Connect<Self> for Interned<dyn DynType> {}

impl Type for Interned<dyn DynCanonicalType> {
    type CanonicalType = Interned<dyn DynCanonicalType>;
    type Value = DynCanonicalValue;
    type CanonicalValue = DynCanonicalValue;
    type MaskType = Interned<dyn DynType>;
    type MaskValue = DynValue;

    impl_match_values_as_self!();

    fn mask_type(&self) -> Self::MaskType {
        (**self).mask_type_dyn()
    }

    fn canonical(&self) -> Self::CanonicalType {
        *self
    }

    fn source_location(&self) -> SourceLocation {
        (**self).source_location_dyn()
    }

    fn type_enum(&self) -> TypeEnum {
        (**self).type_enum_dyn()
    }

    fn from_canonical_type(t: Self::CanonicalType) -> Self {
        t
    }

    fn canonical_dyn_impl(this: &Self) -> Interned<dyn DynCanonicalType> {
        *this
    }

    fn to_dyn_impl(this: &Self) -> Interned<dyn DynType> {
        Interned::cast_unchecked(*this, <dyn DynCanonicalType>::as_dyn_type)
    }

    fn as_dyn_canonical_type_impl(this: &Self) -> Option<&dyn DynCanonicalType> {
        Some(&**this)
    }
}

impl Connect<Self> for Interned<dyn DynCanonicalType> {}

impl sealed::Sealed for Interned<dyn DynCanonicalType> {}

impl CanonicalType for Interned<dyn DynCanonicalType> {
    const CANONICAL_TYPE_KIND: CanonicalTypeKind = CanonicalTypeKind::DynCanonicalType;
}

impl sealed::Sealed for ClockType {}

impl sealed::Sealed for Clock {}

impl sealed::Sealed for AsyncResetType {}

impl sealed::Sealed for AsyncReset {}

impl sealed::Sealed for SyncResetType {}

impl sealed::Sealed for SyncReset {}

impl sealed::Sealed for ResetType {}

impl sealed::Sealed for Reset {}

impl sealed::Sealed for DynBundleType {}

impl sealed::Sealed for DynBundle {}

impl<Signed: GenericConstBool> sealed::Sealed for DynIntType<Signed> {}

impl<Type: IntTypeTrait> sealed::Sealed for IntValue<Type> {}

impl sealed::Sealed for ArrayType<[DynCanonicalValue]> {}

impl sealed::Sealed for Array<[DynCanonicalValue]> {}

impl sealed::Sealed for DynEnumType {}

impl sealed::Sealed for DynEnum {}
